/* OSIALOD written by Mark Spankus */

#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdarg.h>
#if _WIN32
#include <io.h>
#endif


#include "osialod.h"
// usage:
// osialod [-l<address>][-g<address>][-t] <filename>
// -l<address> specify load address
// -g<address> specify execute address
// -a 65A format
// -v 65V format
// -q quickload format
// -c checksum format
// -i<address> include checksum/qloader loader at <address>
// -t test mode; use DOS/WINDOWS End Of Line (not compatible with OSI)
// -mMESSAGE -include MESSAGE before checksum data
// 
 

//65A monitor commands
// R - reset
// L - load mode followed by <2 Byte Address in hex><data as hex>, R to exit
// P - print mode followed by <2 Byte Address in hex> dump data at address, Any key to exit
// G - go with $012D=K, $012A=X, $0129=Y, $012B=A, addr=($012F lo,$012E hi) 
//     via RTI Note that unlike RTS, RTI does NOT add one to the destination before placing 
//     it in the Program Counter (PC)
// 0-9,A-F hex digit
//
// Lxxxx
// 00 11 22 33 44 55 66 77
// 88 99 AA BB CC DD EE FF
// R
// L0120
//


//65V monitor commands
// . switch to address mode
// / switch to data mode
// <CR> increment address (in data mode)
// L set input to serial port. To transfer back, set $00FD to $00 via serial
//  or press reset or SPACE on polled Keyboard
// 0-9,A-F hex digit

/*
The MOS Technology format 
The Lines
Each line consists of 5 fields. These are the length field, address field, data field, 
and the checksum. The lines always start with a semicolon (;) character.

The Fields
    ___ ________ _________ ______ __________ ______
   |_;_|_Length_|_Address_|_Data_|_Checksum_|_CRLF_|
	

Length
The record length field is a 2 character (1 byte) field that specifies the number of 
data bytes in the record. Typically this is 24 or less.
	

Address
This is a 2-byte address that specifies where the data in the record is to be loaded 
into memory, big-endian.
	

Data
The data field contains the executable code, memory-loadable data or descriptive 
information to be transferred.

Checksum
The checksum is an 2-byte field that represents the least significant two bytes of 
the the sum of the values represented by the pairs of characters making up the records 
length, address, and data fields, big-endian.

End of File	(optional?)
The final line should have a data length of zero, and the count of data lines in 
the address field. The checksum is not the usual checksum, it is instead a repeat 
of the data line count.

OSI specific
The final line consists of a $ character followed by 4 digit hex address to start execution
*/



#define BUFSIZE 256
static int LineEnd = 0;
static long load = 0, go = -1, chkldr = -1;
static int pad = 5; //nul padding
static int overwrt = 0;
static int mode = 0;
static int loader = 0;
static char message[80] = {0};
/////////////////////////////////////////////////////////////////////////////
// void main (int argc, char *argv[])
// Synopsis:
//  This is where it all begins
//
int main (int argc, char *argv[])
{


	int test = -1, args, huh = 0;
	char *infilename = NULL;
	char *outfilename = NULL;

	
	FILE *ifp = NULL;
	FILE *ofp = NULL;
	
	if (argc == 1)
	{
		help();
		return 0;
	}
	
	for (args = 1; args < argc; args++)
	{
		if ( *argv[args] == '-')
		{
			char c = *(argv[args]+1);
		    	
		    switch ( tolower(c))
		    {
		    	case 'l':
		    		load = ReadHEXn( argv[args]+2, 4);
		    		break;
		    	case 'g':
		    	case 'h':
		    		go = ReadHEXn( argv[args]+2, 4);
		    		break;
				case 'v':
					mode = 1;
					break;
				case 'a':
					mode = 4;
					break;
				case 'c':
					mode = 2;
					break;
				case 'q':
					mode = 3;
					break;
		    	case 't':
		    		LineEnd = 1;
					if (*(argv[args]+2) == '0')
						LineEnd= -1;
		    		break;
				case 'p':
					pad = ReadHEXn( argv[args]+2, 2);
					break;
				case 'i':
					if (*(argv[args]+2) != 0) //no addr
						chkldr = ReadHEXn( argv[args]+2, 4);
					
					loader=1;
					if (mode == 0)
						mode = 2; //previous behavior
					break;
				case 'f':
					overwrt = 1;
					break;
				case 'm':
					ReadMesg(argv[args]+2, message, sizeof(message));
					
					break;
		    	default:
					huh = 1;
		    		break;
		    }
		}
		else
		{
			if (infilename)
			{
				if (!outfilename)
					outfilename = argv[args];
			}
			else
			{
				infilename = argv[args];
			}
		}
	
	}
	
	if (huh)
	{
		help();
		return 0;
	}
	else
	{
		if (!mode)
		{
			fputs("\nNo action specified. Nothing to do. Use one of -[avciq]\n", stderr);
			return -1;
		}
		
		if (!infilename)
		{
			ifp = stdin;
			//ensure our binary \n's don't convert to \r\n!!!!!
#ifdef _WIN32
			_setmode(fileno(ifp), _O_BINARY);
#endif
		}	
		else
		{
			ifp = fopen(infilename, "rb");
		}
		if (!ifp)
		{
			fputs("\nError opening input file for read.\n", stderr);
			return -1;
		}
	}
	
	if (!outfilename)
	{
		ofp = stdout;
		//ensure our binary \n's don't convert to \r\n!!!!!
#ifdef _WIN32
		_setmode(fileno(stdout), _O_BINARY);
#endif
	}
	else
	{
		//check file exists
		ofp = fopen(outfilename, "rb");
		if (ofp)
		{
			fclose(ofp);
			
			if (!overwrt)
			{
				fprintf(stderr, "Output file %s exists! Use -f to force overwrite.\r\n", outfilename);
				return -1;
			}
	
		}
		
		ofp = fopen(outfilename, "wb");
		if (!ofp)
		{
			fputs("\nError opening output file for write.\n", stderr);
			
			if (ifp && ifp != stdin)
			{
				fclose(ifp);
				ifp = NULL;
			}
			return 1;
		}
	}
	
	switch (mode)
	{
		case 4:
			Write65A( ifp, ofp );
			break;
		case 1:
			Write65V( ifp, ofp );
			break;
		case 2:
			if (loader == 1)
			{
				if (chkldr <= 0)
				{
					chkldr = load - 256;
					if (load < 0x0200)
					   load = 0x1F00;
				}
				WriteChkLoader(	chkldr, ofp);

			}
			WriteChecksum( ifp, ofp );
			break;
		case 3:
			if (loader == 1)
			{
				if (chkldr <= 0)
				{
					chkldr = load - 256;
					if (load < 0x0200)
					   load = 0x1F00;
				}
				WriteQLoader(	chkldr, ofp);
			}
			WriteQSave( ifp, ofp );
			break;

	}

	if (ofp != stdout)
		fclose(ofp);
	if (ifp != stdin)
		fclose(ifp);	

	return 0;
}

//write file in OSI 65A serial monitor load format
int Write65A( FILE* ifp, FILE* ofp )
{

    int nRet = 0, i;
    char buf[BUFSIZE];
    
    WriteALoad(load, ofp);
    	
    while (nRet != EOF)
    {
    	
    	int row = 0;
    	int numRead = fread( buf, sizeof(char), sizeof(buf), ifp);
    	
    	if (!numRead)
    	{
    		break;
    	}
    	
    	for ( i = 0; i < numRead && nRet != EOF; i++)
    	{
    		nRet = WriteHEX(buf[i], ofp);
    		
    		row++;
    		if (row < 8)
    		{
    			//non HEX characters are ignored but make file easier to read for humans
				if (LineEnd >= 0)
					nRet = fputc(' ', ofp);
    		}
    		else
    		{
    			row = 0;
    			nRet = WriteEOL(ofp);
    		}
    		
    	}
    	
    	
    }
    fclose(ifp);
       
    if (nRet != EOF)
    	fputc('R', ofp);
     	
      
    if (go >= 0 && nRet != EOF)
	{
		if (load <= 0x12E ) //no need to write loader
		{			
			//execute
			
			//since we cant pause output for a character count or two, just repeat command
			// a few times, maybe it will catch
			for (i = 0; i < 4 && nRet != EOF; i++)
				nRet = fputc( 'G', ofp); //write nulls for padding after R command
 

		}
		else
		{
			//try to resync ACIA bitstream with some L commands
			for (i = 0; i < 4 && nRet != EOF; i++)
			{
    			nRet = fputc( 'L', ofp);
			}
    		if (nRet != EOF)
				nRet = WriteAGo(go, ofp); 
		}
    }	
    fclose(ofp);
    
    return nRet;
}

//write file data in OSI65V monitor load format
int Write65V( FILE* ifp, FILE* ofp )
{

    int nRet = 0, i;
    char buf[BUFSIZE];
    
    nRet = WriteVLoad(load, ofp); //store address
    nRet = fputc('/', ofp);			//set data mode
    
	while (nRet != EOF)
    {
    	
    	int row = 0;
    	int numRead = fread( buf, sizeof(char), sizeof(buf), ifp);
    	
    	if (!numRead)
    	{
    		break;
    	}
    	
    	for ( i = 0; i < numRead && nRet != EOF; i++)
    	{
    		nRet = WriteHEX(buf[i], ofp);
    		if (nRet != EOF)
				nRet = fputc('\r', ofp);   		
    	}
    	
    	
    }
    fclose(ifp);
       
    if (go >= 0 && nRet != EOF)
    	nRet = WriteVGo(go, ofp); 
    	
    fclose(ofp);
    
    return nRet;
}

//Write file data in MOS/OSI checksum format
//allowed characters for .chk are
//;0-9, A-G, <nul> <lf> <cr>
//;1808004C2808A9D08D1508A9E08D1A08A200A008A9209D00D0A9080928<cr>
//;<len hex2><adrhi hex2><adrlo hex2><data hex2><chk hi hex2><chk lo hex2><cr><lf>
int WriteChecksum( FILE* ifp, FILE* ofp )
{
	unsigned long addr = load;

    int nRet = 0, i;
	unsigned short sum;
    char buf[BUFSIZE];
       	
    if (*message)
	{
		nRet = fwrite(message, 1, strlen(message), ofp);
	}
	while (nRet != EOF)
    {
    	
    	int maxrow = 0x18;
    	int numRead = fread( buf, sizeof(char), maxrow, ifp);
    	
    	if (!numRead)
    	{
    		break;
    	}
		//addr<data><cksm><CR>
		nRet = fputc(';', ofp);
		
		sum = numRead;
		if (nRet != EOF)
			nRet = WriteHEX( (unsigned char)numRead, ofp );
		
		if (nRet != EOF)
		{
			sum += (unsigned char)((addr >> 8) & 0xFF);
			//write current addr
			if (nRet != EOF)
				nRet = WriteHEX( (unsigned char)((addr >> 8) & 0xFF), ofp );

			sum += (unsigned char)(addr & 0xFF);
			if (nRet != EOF)
				nRet = WriteHEX( (unsigned char)(addr & 0xFF), ofp );
		}

		addr += numRead;
    	
    	for ( i = 0; i < numRead && nRet != EOF; i++)
    	{
    		nRet = WriteHEX(buf[i], ofp);
    		sum += (unsigned char)buf[i];
		}
		if (nRet != EOF)
			nRet = WriteHEX( (unsigned char) ((sum >> 8) & 0xFF), ofp);
		
		if (nRet != EOF)
			nRet = WriteHEX( (unsigned char) (sum  & 0xFF), ofp);

    	//<CR> ends data
		if (nRet != EOF)
			fputc('\r', ofp);
		
		//padding?
		for ( i = 0; i < pad && nRet != EOF; i++)
		{
			nRet = fputc( '\x0', ofp);
		}
		//<LF> starts next line
		if (nRet != EOF)
			fputc('\n', ofp);
    	
    	
    }
    fclose(ifp);
     
    if (go >= 0 && nRet != EOF)
	{
		if (nRet != EOF)
			nRet = fputc( '$', ofp); //set execute addr	mode
			
		//write hex
		if (nRet != EOF)
			nRet = WriteHEX( (unsigned char)((load >> 8) & 0xFF), ofp );
		if (nRet != EOF)
			nRet = WriteHEX( (unsigned char)(load & 0xFF), ofp ); 
    }	
    fclose(ofp);
    
    return nRet;
}


//8bit version of MOS/OSI checksum format
//qsave *.qsv
// ;<len byte><adrh byte><adrl byte><len data byte><chksum of whole line after ; hi><chksm lo>
//$<adrhi><adrlo> [EXECUTE]
int WriteQSave( FILE* ifp, FILE* ofp )
{
	unsigned long addr = load;

    int nRet = 0, i;
	unsigned short sum;
    char buf[BUFSIZE];
      	
    if (*message)
	{
		nRet = fwrite(message, 1, strlen(message), ofp);
	}
    
	while (nRet != EOF)
    {
    	
    	int maxrow = 0x38;
    	int numRead = fread( buf, sizeof(char), maxrow, ifp);
    	
    	if (!numRead)
    	{
    		break;
    	}
		//addr<data><cksm><CR>
		nRet = fputc(';', ofp);
		
		sum = numRead;
		if (nRet != EOF)
			nRet = fputc((unsigned char)numRead, ofp);
		
		if (nRet != EOF)
		{
			sum += (unsigned char)((addr >> 8) & 0xFF);
			//write current addr
			if (nRet != EOF)
				nRet = fputc((unsigned char)((addr >> 8) & 0xFF), ofp);


			sum += (unsigned char)(addr & 0xFF);
			if (nRet != EOF)
				nRet = fputc( (unsigned char)(addr & 0xFF), ofp );
		}

		addr += numRead;
    	
    	for ( i = 0; i < numRead && nRet != EOF; i++)
    	{
    		nRet = fputc((unsigned char)buf[i], ofp);
    		sum += (unsigned char)buf[i];
		}
		if (nRet != EOF)
			nRet = fputc( (unsigned char) ((sum >> 8) & 0xFF), ofp);
		
		if (nRet != EOF)
			nRet = fputc( (unsigned char) (sum  & 0xFF), ofp);  	
    	
    }
    fclose(ifp);
     
    if (go >= 0 && nRet != EOF)
    	nRet = WriteQGo(go, ofp); 
    	
    fclose(ofp);
    
    return nRet;
}

int ReadMesg(const char *src, char *dest, int nMax)
{
	int nSize = 0;
	bool backslash = false;
	while (src && *src && nSize < nMax-1)
	{
		char c = *src++;
		if (c && c != ';' || c != '$')
		{
			if (c == '\\')
			{
				if (backslash)
					backslash = false;
				else
				{
					backslash = true;
					continue;
				}
			}
			else if (backslash)
			{
				switch (tolower(c))
				{
					case 'r':
						c = '\r';
						break;

					case 'n':
						c = '\n';
						break;

					case 'b':
						c = '\b';
						break;

					case 'f':
						c = '\f';
						break;

					case 't':
						c = '\t';
						break;
					default:
						break;
				}
				backslash = false;
			}
			dest[nSize++] = c;
		}
		if (!c || c == '"')
			break;
	}
	dest[nSize] = 0;
	return nSize;
}

long ReadHEXn(const char *pstr, int width)
{
    
    long nRet = 0;
    
	short count = 0;
	
	if (!pstr || width < 1 || width > 8)
       return 0;
       
    while (*pstr && count < width)
    {
    
    	char c = *pstr;
    	
    	if ((c == 0x20 || c == 0x0A || c == 0x0D || c == 0x09) && count == 0)
    	{
    		pstr++;
			continue;
		}
    	
    	nRet = nRet << 4;
    	
    	if (c >= 'a' && c <= 'f')
    		c -= 0x20;
    		
    	if (c >= 'A' && c <= 'F')
    		nRet |= (c-'7');
    	else if (c >= '0' && c <= '9')
    		nRet |= (c-'0');
    	else
            break;
        pstr++;
        count++;
    
    }
    
    return nRet;
	
}

int  WriteHEX( unsigned char data, FILE *out)
{
	unsigned char c;
	int nRet = 0;
	
	c = data >> 4;
	if (c > 9 )
		c = c + 7;
		
	nRet = fputc( c+'0', out);
	c = data & '\x0F';
	
	if (c > 9)
		c += 7;
		
	if (nRet != EOF)
		nRet = fputc( c+'0', out);
		
	return nRet;
	

}

//write LOAD addr for 65A
int WriteALoad(long load, FILE* out)
{
	int nRet = 0;
	
	if (load < 0)
		return 0;
		
	nRet = fputc( 'L', out);
		
	if (nRet != EOF)
		nRet = WriteHEX( (unsigned char)((load >> 8) & 0xFF), out );
	if (nRet != EOF)
		nRet = WriteHEX( (unsigned char)(load & 0xFF), out );
	if (nRet != EOF)
		nRet = WriteEOL( out );
	
	return nRet;
} 

//write LOAD addr for 65V
int WriteVLoad(long load, FILE* out)
{
	int nRet = 0;
	
	if (load < 0)
		return 0;
		

	nRet = fputc( '.', out); //set address mode
		
	//write hex
	if (nRet != EOF)
		nRet = WriteHEX( (unsigned char)((load >> 8) & 0xFF), out );
	if (nRet != EOF)
		nRet = WriteHEX( (unsigned char)(load & 0xFF), out );
	
	return nRet;
}

// Upon G command, Serial monitor loads from the following
//
// $0129 - Y
// $012A - X
// $012B - A
// $012C - Proc Status  'P'
// $012D - stack pointer 'K'
// $012E - PC Hi
// $012F - PC Lo
// When in Command Mode. 65A resets ACIA on every character except
// L, P, & G
//
int WriteAGo(long addr, FILE* out)
{
    
    int i, nRet = 0;
	//load address 0129, Y=0, X=0, A=0, P=4, k=FD
	char szLoad[] = {"00000004FD"};
    
	if (addr < 0)
		return 0;

		
	nRet = WriteALoad( 0x0129, out);  //store address, clear registers
	
	for (i=0; i< (sizeof(szLoad)/sizeof(char))-1 && nRet != EOF; i++)
		nRet = fputc(szLoad[i], out);
	
	if (nRet != EOF)
		WriteHEX( (unsigned char)((addr >> 8) & 0xFF), out );
		
	if (nRet != EOF)
		WriteHEX( (unsigned char)(addr & 0xFF), out );
	
	if (nRet != EOF)
		nRet = fputc('R', out);

	//the 'R' command caused the ACIA to be reset, along with reinitialization of the
	//65A monitor.  Some characters are lost from the data stream and serial syncronization 
	//may be compromised.  Anything other than P,G,L will also reset ACIA
			
	//since we cant pause output for a character count or two send a few NULS then write G
	for (i = 0; i < 3 && nRet != EOF; i++)
		nRet = fputc( '\0', out);
	
	if (nRet != EOF)
		fputc( 'G', out); 
	
	return nRet;


}

//Write go for 65V
int WriteVGo(long addr, FILE* out)
{
    
    int nRet = 0;
    
	if (addr < 0)
		return 0;
		
	nRet = WriteVLoad( addr, out);  //store address
	
	if (nRet != EOF)
		nRet = fputc( 'G', out);     //execute address
	
	return nRet;


}
//Write go for MOS/OSI checksum loader
int WriteCGo(long addr, FILE* out)
{
    
    int nRet = 0;
    
	if (addr < 0)
		return 0;

	nRet = fputc( '$', out); //set address mode
		
	//write hex
	if (nRet != EOF)
		nRet = WriteHEX( (unsigned char)((addr >> 8) & 0xFF), out );
	if (nRet != EOF)
		nRet = WriteHEX( (unsigned char)(addr & 0xFF), out );
		

	return nRet;
}

//Write go for QSave
int WriteQGo(long addr, FILE* out)
{
    
    int nRet = 0;
    
	if (addr < 0)
		return 0;

	nRet = fputc( '$', out); //set address mode
		
	//write addr
	if (nRet != EOF)
		nRet = fputc( (unsigned char)((addr >> 8) & 0xFF), out);
	if (nRet != EOF)
		nRet = fputc( (unsigned char)(addr & 0xFF), out);

	return nRet;
}

//write end of line for OSI or PC
int WriteEOL(FILE *out)
{

	switch(LineEnd)
	{
		case -1:
			return 0;
			break;
		
		case 0:
			return fputc('\x0D', out);
			break;
			
		default:
        case 1:
			return fputs("\x0D\x0A", out);
			break;
		
	}
    
}


void help()
{
	char buf[] =
	{
        "\r\nV1.08 Usage:\r\n\r\nOSIALOD [-l<hex address>][-g<hex address>][-t][-f][-p<nn>]\r\n        [-a|-v|-i<hex addr>|-c|-q] [InFile] [OutFile]\r\n\r\n"
		"-a  Convert binary data to OSI 65A Serial Monitor Load Format\r\n"
		"-v  Convert binary data to OSI 65V ROM Monitor Load Format\r\n"
		"-c  Convert binary data to MOS/OSI Checksum Load Format\r\n"
		"-q  Convert binary data to OSI QSave Binary Checksum Load Format\r\n"
        "-i[addr]  Convert data to hybrid 65V / Checksum/QSave load format for BASIC-IN-ROM OSIs\r\n" 
		"        (optional address specifies loader location. Defaults to load addr-256)\r\n"
		"-g<addr>  go/execute address; Defaults to load address.\r\n"
		"-l<addr>  load/target address of file\r\n"
		"-p<nn> nn=number of bytes of nul padding at end of line in hex) defaults to 5\r\n"
		"-m\"Message\" Add a text message before the start of checksum/qsave format\r\n"
		"-f  Force overwrite of target file\r\n"
		"-t  Use CR/LF End of Line. Normally just CR\r\n"
		" Infile can be redirected console input (stdin), Outfile defaults to (stdout)\r\n\r\n"
		" ex: OSIAlod -i -l0300 6502.bin file.lod\r\n"
		"     OSIAlod -i0300 > lod.chk < nul  (emit checksum loader at $0300 with no data)\r\n"

    };
    fputs(buf, stderr);
}

//Checksum loader data  0EF data LOADERDATA  
unsigned char LOADERDATA[] = {
 0x00, 0x00, 0x00, 0x00, 0x00, 0x20, 0xA2, 0x00, 0xC9, 0x3B, 0xF0, 0x13, 0xC9, 0x24, 0xD0, 0xF5,
 0x20, 0x83, 0x00, 0x8D, 0x03, 0x00, 0x20, 0x83, 0x00, 0x8D, 0x02, 0x00, 0x4C, 0xB0, 0x00, 0xA9,
 0x00, 0x8D, 0x02, 0x00, 0x8D, 0x03, 0x00, 0x20, 0x70, 0x00, 0xAA, 0xF0, 0xD8, 0x20, 0x70, 0x00,
 0x8D, 0x3E, 0x00, 0x20, 0x70, 0x00, 0x8D, 0x3D, 0x00, 0x20, 0x70, 0x00, 0x8D, 0xDE, 0xAD, 0xEE,
 0x3D, 0x00, 0xD0, 0x03, 0xEE, 0x3E, 0x00, 0xCA, 0xD0, 0xEF, 0x20, 0x83, 0x00, 0xCD, 0x03, 0x00,
 0xD0, 0x08, 0x20, 0x83, 0x00, 0xCD, 0x02, 0x00, 0xF0, 0xAB, 0xA0, 0x00, 0xB9, 0xB8, 0x00, 0xF0,
 0x06, 0x20, 0xEE, 0xFF, 0xC8, 0xD0, 0xF5, 0x20, 0xA5, 0x00, 0xC9, 0x47, 0xD0, 0xF9, 0xF0, 0x95,
 0x20, 0x83, 0x00, 0x18, 0x6D, 0x02, 0x00, 0x8D, 0x02, 0x00, 0x90, 0x03, 0xEE, 0x03, 0x00, 0xAD,
 0x04, 0x00, 0x60, 0x20, 0x86, 0x00, 0x20, 0xA2, 0x00, 0xC9, 0x41, 0x90, 0x02, 0xE9, 0x07, 0x29,
 0x0F, 0x0A, 0x0A, 0x0A, 0x0A, 0xA0, 0x04, 0x2A, 0x2E, 0x04, 0x00, 0x88, 0xD0, 0xF9, 0xAD, 0x04,
 0x00, 0x60, 0xA9, 0x80, 0x2C, 0xA9, 0x00, 0x8D, 0x03, 0x02, 0x20, 0xEB, 0xFF, 0x4C, 0xEE, 0xFF,
 0xA9, 0x00, 0x8D, 0x03, 0x02, 0x6C, 0x02, 0x00, 0x0D, 0x0A, 0x0A, 0x4F, 0x42, 0x4A, 0x20, 0x43,
 0x48, 0x45, 0x43, 0x4B, 0x53, 0x55, 0x4D, 0x20, 0x45, 0x52, 0x52, 0x0D, 0x0A, 0x52, 0x45, 0x57,
 0x49, 0x4E, 0x44, 0x20, 0x50, 0x41, 0x53, 0x54, 0x20, 0x45, 0x52, 0x52, 0x20, 0x2D, 0x20, 0x54,
 0x59, 0x50, 0x45, 0x20, 0x47, 0x20, 0x54, 0x4F, 0x20, 0x52, 0x45, 0x53, 0x54, 0x41, 0x52, 0x54,
 0x00 };

//base is $0000 /start is $0005																		  
// Relocation offsets 
 unsigned char relocOFFSETS[] = {
0x06, 0x11, 0x14, 0x17, 0x1A, 0x1D, 0x22, 0x25,                 
0x28, 0x2E, 0x31, 0x34, 0x37, 0x3A, 0x40, 0x45,                 
0x4B, 0x4E, 0x53, 0x56, 0x5D, 0x68, 0x71, 0x75,                 
0x78, 0x7D, 0x80, 0x84, 0x87, 0x99, 0x9F, 0xB6 };

/* Embedded file: qload.bin */
const unsigned char QLOADDATA[] = {
0xd8,0x78,0xa2,0x03,0x20,0xc8,0x00,0xae,0xe2,0xff,0xca,0x86,0xf4,0x20,0xba,0x00,
0xc9,0x24,0xf0,0x4e,0xc9,0x3b,0xd0,0xf5,0xa0,0x00,0x84,0xf0,0x84,0xf1,0x20,0x80,
0x00,0xaa,0xe0,0x39,0xb0,0xe7,0x20,0x80,0x00,0x85,0xf3,0x20,0x9e,0x00,0x20,0x80,
0x00,0x85,0xf2,0x20,0x9e,0x00,0x20,0x80,0x00,0x91,0xf2,0xc8,0xca,0xd0,0xf7,0x20,

0x8f,0x00,0xc5,0xf1,0xd0,0x0e,0x20,0x8f,0x00,0xc5,0xf0,0xd0,0x07,0xa2,0x00,0x20,
0xc8,0x00,0xf0,0xb9,0xa2,0x06,0x20,0xc8,0x00,0x20,0xbd,0x00,0xc9,0x47,0xd0,0xf9,
0xf0,0xab,0xa2,0x00,0x20,0x8f,0x00,0x85,0xf1,0x20,0x8f,0x00,0x85,0xf0,0xa2,0x14,
0x20,0xc8,0x00,0xa5,0xf1,0x20,0x9e,0x00,0xa5,0xf0,0x20,0x9e,0x00,0x6c,0xf0,0x00,

0x20,0x8f,0x00,0x48,0x18,0x65,0xf0,0x85,0xf0,0x90,0x02,0xe6,0xf1,0x68,0x60,0x20,
0xeb,0xff,0x24,0xf4,0x30,0x04,0x8d,0x45,0xd7,0x60,0x8d,0x6a,0xd3,0x60,0x48,0x48,
0x4a,0x4a,0x4a,0x4a,0x20,0xad,0x00,0x68,0x20,0xad,0x00,0x68,0x60,0x29,0x0f,0x09,
0x30,0xc9,0x3a,0x90,0x02,0x69,0x06,0x4c,0xee,0xff,0xa9,0x80,0x2c,0xa9,0x00,0x8d,

0x03,0x02,0x20,0xeb,0xff,0x4c,0xee,0xff,0xbd,0xd4,0x00,0xf0,0x06,0x20,0xee,0xff,
0xe8,0xd0,0xf5,0x60,0x20,0x4f,0x4b,0x0a,0x0d,0x00,0x20,0x45,0x52,0x52,0x2d,0x54,
0x59,0x50,0x45,0x20,0x47,0x0a,0x0d,0x00,0x0a,0x0d,0x4a,0x55,0x4d,0x50,0x20,0x24,
0x00 };

//base is $0000 /start is $0000																		  
// Relocation offsets 
unsigned char QrelocOFFSETS[] = {
0x05,0x0E,0x1F,0x27,0x2c,0x2F,0x34,0x37, 
0x40,0x47,0x50,0x57,0x5A,0x65,0x6A,0x71,
0x76,0x7B,0x81,0xA5,0xA9,0xC9   
 };


/* 
   ; FOR BASIC-IN-ROM MACHINES
OUTCHR = $FFEE
INCHR  = $FFEB	 
STOR =*                     
		.BYTE 0,0,0,0,0
START               JSR LOADON
                    CMP #$3B	;START OF LINE CHAR
                    BEQ RDCHK
                    CMP #'$
                    BNE START
                    JSR GETHEX
                    STA STOR+3
                    JSR GETHEX
                    STA STOR+2
                    JMP EXEC
                    
RDCHK               LDA #$00
                    STA STOR+2
                    STA STOR+3   ;INIT CHKSUM
                    JSR ADDCHK
                    TAX			
                    BEQ START ; ignore line
                    JSR ADDCHK
                    STA ADDRHI
                    JSR ADDCHK
                    STA ADDRLO
STORIT              JSR ADDCHK
ADDRLO =*+1
ADDRHI =*+2
                    STA $ADDE
                    INC ADDRLO
                    BNE NOINC
                    INC ADDRHI
NOINC               DEX
                    BNE STORIT
                    JSR GETHEX
                    CMP STOR+3
                    BNE PRINT
                    JSR GETHEX
                    CMP STOR+2
                    BEQ START
PRINT               LDY #$00
MLOOP               LDA MESG,Y
                    BEQ WAITG
                    JSR OUTCHR
                    INY
                    BNE MLOOP
WAITG               JSR LOADOFF
                    CMP #'G
                    BNE WAITG
                    BEQ START
ADDCHK              JSR GETHEX
                    CLC
                    ADC STOR+2
                    STA STOR+2
                    BCC ADDCH1
                    INC STOR+3
ADDCH1              LDA STOR+4
                    RTS
                    
GETHEX              JSR GETNIB
GETNIB              JSR LOADON
                    CMP #$41
                    BCC NOTHI
                    SBC #$07
NOTHI               AND #$0F
                    ASL A
                    ASL A
                    ASL A
                    ASL A
                    LDY #$04
HEXROL              ROL A
                    ROL STOR+4
                    DEY
                    BNE HEXROL
                    LDA STOR+4
                    RTS
                    
LOADON
					LDA #$80
LOADOFF = *+1
					.BYTE $2C, $A9,  $00 ;BIT $00A9 FORCE NO ZERO PAGE 
                    STA $0203            ;FLAG TO ENABLE INPUT FROM SERIAL
                    JSR INCHR
                    JMP OUTCHR
                    
EXEC                LDA #$00
                    STA $0203
                    JMP (STOR+2)
MESG
					.BYTE $D,$A,$A,'OBJ CHECKSUM ERR'
					.BYTE $D,$A,'REWIND PAST ERR - TYPE G TO RESTART',0
;END
*/

/*
binary checksum loader for BASIC-In-ROM machines
; Quick Loader  Mark Spankus 08/83
SCNCHR=$D745
SCRNC1=$D36A
STORE =$F2
CHKSUM=$F0
FLAG  =$F4

OUTCHR = $FFEE
INCHR  = $FFEB

	*=$1500
	.EXE *
	CLD
	SEI
	LDX #NLIDX
	JSR MESGPRINT
	LDX $FFE2	;basic rom 1K/2K flag
	DEX
	STX FLAG
START
	JSR LOADON
	CMP #$24
	BEQ JUMP
	CMP #$3B
	BNE START
	LDY #$00
	STY CHKSUM
	STY CHKSUM+1
	JSR ADDGET
	TAX
	CPX #$39	;check length for expected max
	BCS START
	JSR ADDGET
	STA STORE+1
	JSR HEXOUT
	JSR ADDGET
	STA STORE
	JSR HEXOUT
LOOP 
	JSR ADDGET
	STA (STORE),Y
	INY
	DEX
	BNE LOOP
	JSR GET
	CMP CHKSUM+1
	BNE ERR
	JSR GET
	CMP CHKSUM
	BNE ERR
OK
	LDX #OKIDX
	JSR MESGPRINT
	BEQ START
ERR 
	LDX #ERRIDX
	JSR MESGPRINT
WAITG
    JSR LOADOFF
	CMP #'G
	BNE WAITG
	BEQ START
JUMP 
	LDX #$00
	JSR GET
	STA CHKSUM+1
	JSR GET
	STA CHKSUM
	LDX #GOIDX
	JSR MESGPRINT
	LDA CHKSUM+1
	JSR HEXOUT
	LDA CHKSUM
	JSR HEXOUT
	JMP (CHKSUM)
ADDGET
	JSR GET
	PHA
	CLC
	ADC CHKSUM
	STA CHKSUM
	BCC *+4
	INC CHKSUM+1
	PLA
	RTS
GET
	JSR INCHR
	BIT FLAG
	BMI C1DISP
	STA SCNCHR
	RTS
C1DISP
	STA SCRNC1
	RTS
HEXOUT
	PHA
	PHA
	LSR A
	LSR A
	LSR A
	LSR A
	JSR OUT
	PLA
	JSR OUT
	PLA
	RTS
OUT
	AND #$0F
	ORA #$30
	CMP #$3A
	BCC *+4
	ADC #$06
	JMP OUTCHR
LOADON
	LDA #$80
LOADOFF = *+1
	.BYTE $2C, $A9, $00 ;BIT $00A9 FORCE NO ZERO PAGE 
	STA $0203  ;FLAG TO ENABLE INPUT FROM SERIAL
	JSR INCHR
	JMP OUTCHR
	
MESGPRINT
	LDA TEXTOK,X
	BEQ TXTDONE
	JSR OUTCHR
	INX
	BNE MESGPRINT
TXTDONE
	RTS
TEXTOK=*
	.BYTE ' OK',$A,$D,$0
TXTNL=*-3
TEXTERR=*
	.BYTE ' ERR-TYPE G',$A,$D,$0
TEXTGO=*
	.BYTE $A,$D,'JUMP $',0
OKIDX=0
ERRIDX=TEXTERR-TEXTOK
GOIDX =TEXTGO-TEXTOK
NLIDX=TXTNL-TEXTOK 
*/


//reloc offset to 2 bytes @ offset

int WriteChkLoader(int nTargetAddr, FILE* out)
{
	int nRet = EOF, i;
	unsigned char* pBuf = (unsigned char*)malloc( sizeof(LOADERDATA) + 1);
	if (pBuf)
	{
		//patch chk loader with reloc offsets
		memcpy(pBuf, LOADERDATA, sizeof(LOADERDATA) );

		for ( i = 0; i < sizeof(relocOFFSETS); i++)
		{
			unsigned char *p = pBuf + relocOFFSETS[i];
			short addr = *p + (*(p+1) << 8);
			addr += nTargetAddr;
			*p = addr & 0xFF;
			*(p+1) = addr >> 8;
		}

		nRet = WriteVLoad(nTargetAddr, out); //store address
		nRet = fputc('/', out);			//set data mode


    	//write chk loader body
		for ( i = 0; i < sizeof(LOADERDATA) && nRet != EOF; i++)
    	{
    		nRet = WriteHEX(pBuf[i], out);
    		if (nRet != EOF)
				nRet = fputc('\r', out);   		
    	}
		
		if (nRet != EOF)
			nRet = WriteVGo(nTargetAddr+5, out);

		free(pBuf);
	}
	return nRet;

}

int WriteQLoader(int nTargetAddr, FILE* out)
{
	int nRet = EOF, i;
	unsigned char* pBuf = (unsigned char*)malloc( sizeof(QLOADDATA) + 1);
	if (pBuf)
	{
		//patch chk loader with reloc offsets
		memcpy(pBuf, QLOADDATA, sizeof(QLOADDATA) );

		for ( i = 0; i < sizeof(QrelocOFFSETS); i++)
		{
			unsigned char *p = pBuf + QrelocOFFSETS[i];
			short addr = *p + (*(p+1) << 8);
			addr += nTargetAddr;
			*p = addr & 0xFF;
			*(p+1) = addr >> 8;
		}

		nRet = WriteVLoad(nTargetAddr, out); //store address
		nRet = fputc('/', out);			//set data mode


    	//write Q loader body
		for ( i = 0; i < sizeof(QLOADDATA) && nRet != EOF; i++)
    	{
    		nRet = WriteHEX(pBuf[i], out);
    		if (nRet != EOF)
				nRet = fputc('\r', out);   		
    	}
		
		if (nRet != EOF)
			nRet = WriteVGo(nTargetAddr, out);

		free(pBuf);
	}
	return nRet;

}
